/* GPGPU computation of jacobi algorithm with SINGLE PRECISION!
* 
******************************************************************
* Subroutine Jacobi
* Solves poisson equation on rectangular grid assuming : 
* (1) Uniform discretization in each direction, and 
* (2) Dirichlect boundary conditions 
* 
* Jacobi method is used in this routine 
*
* Input : n,m   Number of grid points in the X/Y directions 
*         dx,dy Grid spacing in the X/Y directions 
*         alpha Helmholtz eqn. coefficient 
*         omega Relaxation factor 
*         f(n,m) Right hand side function 
*         u(n,m) Dependent variable/Solution
*         tolerance Tolerance for iterative solver 
*         maxit  Maximum number of iterations 
*
* Output : u(n,m) - Solution 
*****************************************************************
*/
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "jacobi.cuh"

#define THREADSPERBLOCK 256

#define U(j,i) u[(j) * cols + (i)]
#define F(j,i) f[(j) * cols + (i)]
#define UOLD(j,i) uold[(j) * cols + (i)]

/* Performs a single Jacobi iteration on the GPU
* --------------------------------------------------------
* Outer loop has to be distributed among the blocks
* Inner loop has to be distributed among the threads within the block
*/

// This is the GPU kernel
__global__
void oneJacobiIteration(float* u, float* uold, float* f, int rows, int cols, float ax, float ay, float b, float fRelax, float* residuals) { 
	int j = blockIdx.x;
	int t = threadIdx.x;

	// TODO: Create a static array in shared memory of size 'THREADSPERBLOCK'
	float resBlock[THREADSPERBLOCK];
	
	float res = 0.0f;

	// TODO: Verfiy that the computation is only done within the inner elements of the grid
	// TODO: Since THREADSPERBLOCK < cols, one thread has to do compute several elements
	// The best way is to run through a loop in size of THREADSPERBLOCK 
	if (             ) {



			float fLRes = ( ax * (UOLD(j, i-1) + UOLD(j, i+1))
			+ ay * (UOLD(j-1, i) + UOLD(j+1, i))
			+  b * UOLD(j, i) - F(j, i)) / b;
			U(j,i) = UOLD(j,i) - fRelax * fLRes;
			res += fLRes * fLRes;
		}
		
		// TODO: Store the interim residual in the array in shared memory
	
	
		// TODO: Snychronize the threads within the block


		// TODO: Reduce the values in the 'resBlock' (shared mem) to the 'residuals' array in global mem
		// This should only by done by ONE thread within the block




	} else {
		// Initializes the residuals array for the boundary elements
		if (t==0) residuals[j] = 0.0f;
	}	
}


void Jacobi (int iRows, int iCols, int iRowLast, int iIterMax, float fAlpha, float fRelax, float fTolerance, float fDx, float fDy, float *afU, float *afF, float* fResidual, int* iIterCount)
{
	cudaError_t err;
	
	float ax, ay, b, residual = 0.0f;
	int n = iCols * iRows;
	
	float *h_resArray = (float*)malloc(iRows * sizeof(float)); // array to store interim residuals
	float *d_uold, *d_u, *d_f, *d_resArray; // pointer to device
	
	

	// Allocates memory on the device for u, uold and f
	err = cudaMalloc((void**)&d_uold, n * sizeof(float));
	checkErr(err, __FILE__, __LINE__);
	err = cudaMalloc((void**)&d_u, n * sizeof(float));
	checkErr(err, __FILE__, __LINE__);
	err = cudaMalloc((void**)&d_f, n * sizeof(float));
	checkErr(err, __FILE__, __LINE__);

	// Allocates memory on the device to store iterim residuals
	err = cudaMalloc((void**)&d_resArray, iRows * sizeof(float));
	checkErr(err, __FILE__, __LINE__);


	ax = 1.0f / (fDx * fDx);      /* X-direction coef */
	ay = 1.0f / (fDy * fDy);      /* Y_direction coef */
	b = -2.0f * (ax + ay) - fAlpha;     /* Central coeff */
	residual = 10.0f * fTolerance;

	// Copies data from host to device
	err = cudaMemcpy(d_u, afU, n*sizeof(float), cudaMemcpyHostToDevice);
	checkErr(err, __FILE__, __LINE__);
	err = cudaMemcpy(d_uold, afU, n*sizeof(float), cudaMemcpyHostToDevice);
	checkErr(err, __FILE__, __LINE__);
	err = cudaMemcpy(d_f, afF, n*sizeof(float), cudaMemcpyHostToDevice);
	checkErr(err, __FILE__, __LINE__);

	while (*iIterCount < iIterMax /*&& residual > fTolerance*/) 
	{
		residual = 0.0f; 
		
		// swap u and uold
		float *utmp;
		utmp = d_uold;
		d_uold = d_u;
		d_u = utmp;
		
		// TODO: Specify the number of 'blocks_per_grid'
		dim3 threads_per_block(THREADSPERBLOCK);
		dim3 blocks_per_grid(            );
		
		// Launches the kernel
		oneJacobiIteration<<<blocks_per_grid,threads_per_block>>>(d_u, d_uold, d_f, iRows, iCols, ax, ay, b, fRelax, d_resArray);

		// Copies the iterim residuals from the device to the host to finally reduce them on the host
		err = cudaMemcpy(h_resArray, d_resArray, iRows*sizeof(float), cudaMemcpyDeviceToHost);
		checkErr(err, __FILE__, __LINE__);
		
		residual = reduceOnCPU(h_resArray, iRows);

		(*iIterCount)++;
		residual = sqrtf(residual) / (iCols * iRows);
	} // while 
	
	*fResidual = residual;
	
	// Copies final solution u from device to host
	err = cudaMemcpy(afU, d_u, n*sizeof(float), cudaMemcpyDeviceToHost);
	checkErr(err, __FILE__, __LINE__);
	
	// Frees all allocated memory on the device
	cudaFree(d_uold);
	cudaFree(d_u);
	cudaFree(d_f);
	cudaFree(d_resArray);
	
	free(h_resArray);

}

float reduceOnCPU(float *array, int size) {
	int i;
	float sum = 0.0f;
	for (i=0; i<size; i++) {
		sum+=array[i];
	}
	return sum;
}

// Checks whether a CUDA error occured
// If so, the error message is printed and the program exits
inline void checkErr(cudaError_t err, const char* file, const int line)
{
	if(cudaSuccess != err)
	{
		fprintf(stderr, "%s: Cuda error in line %d: %s.\n", file, line, cudaGetErrorString(err) );
		exit(-1);
	}
}

