/* Serial CPU computation of jacobi algorithm with SINGLE PRECISION!
Some small code changes to help the compiler vectorize the inner main loop.
*/

/* 
******************************************************************
* Subroutine HelmholtzJ
* Solves poisson equation on rectangular grid assuming : 
* (1) Uniform discretization in each direction, and 
* (2) Dirichlect boundary conditions 
* 
* Jacobi method is used in this routine 
*
* Input : n,m   Number of grid points in the X/Y directions 
*         dx,dy Grid spacing in the X/Y directions 
*         alpha Helmholtz eqn. coefficient 
*         omega Relaxation factor 
*         f(n,m) Right hand side function 
*         u(n,m) Dependent variable/Solution
*         tolerance Tolerance for iterative solver 
*         maxit  Maximum number of iterations 
*
* Output : u(n,m) - Solution 
*****************************************************************
*/
#if defined(WIN32) || defined (_CLUSTER_OPENMP)
/* fix a visualstudio 2005 bug */
#include <omp.h>
#endif
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "jacobi.h"

#define U(j,i) afU[(j) * data->iCols + (i)]
#define F(j,i) afF[(j) * data->iCols + (i)]
#define UOLD(j,i) uold[(j) * data->iCols + (i)]

void Jacobi (struct JacobiData *data)
{
	/* use local pointers for performance reasons */
	float *afU, *afF;
	int i, j;
	float fLRes;
	
	float ax, ay, b, residual;
	
	float *uold = (float*)malloc(data->iCols * data->iRows * sizeof(float));
	afU = data->afU;
	afF = data->afF;
	
	if (uold)
	{
		ax = 1.0f / (data->fDx * data->fDx);      /* X-direction coef */
		ay = 1.0f / (data->fDy * data->fDy);      /* Y_direction coef */
		b = -2.0f * (ax + ay) - data->fAlpha;     /* Central coeff */
		residual = 10.0f * data->fTolerance;

		while (data->iIterCount < data->iIterMax && residual > data->fTolerance) 
		{
			residual = 0.0f;
			
#pragma omp parallel
			{
				/* copy new solution into old */
#pragma omp for private(j, i)
				for (j = 0; j < data->iRows; j++)
				{
					for (i = 0; i < data->iCols; i++)
					{
						UOLD(j,i) = U(j,i);
					}
				}


				/* compute stencil, residual and update */
#pragma omp for private(j, i, fLRes) reduction(+:residual)
				for (j = 1; j <= data->iRows - 2; j++)
				{
					for (i = 1; i <= data->iCols - 2; i++)
					{
						fLRes = ( ax * (UOLD(j, i-1) + UOLD(j, i+1))
						+ ay * (UOLD(j-1, i) + UOLD(j+1, i))
						+  b * UOLD(j, i) - F(j, i)) / b;

						/* update solution */
						U(j,i) = UOLD(j,i) - data->fRelax * fLRes;

						/* accumulate residual error */
						residual += fLRes * fLRes;
					}
				}
			}

			/* error check */
			(data->iIterCount)++;
			residual = sqrtf(residual) / (data->iCols * data->iRows);
		} /* while */

		data->fResidual = residual;
		free(uold);
	}
	else 
	{
		fprintf(stderr,"Error: can't allocate memory\n");
		Finish(data);
		exit(1);
	}
}

