#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include "realtime.h"
#include "jacobi.h"

#ifdef _ACCEL
#include "accel.h"
#endif

#define U(j,i) data->afU[((j) - data->iRowFirst) * data->iCols + (i)]
#define F(j,i) data->afF[((j) - data->iRowFirst) * data->iCols + (i)]

#ifdef _ACCEL
static void InitGPU(int argc, char** argv) {
	// gets the device id (if specified) to run on
	int devId = -1;
	if (argc > 1) {
		devId = atoi(argv[1]);
		int devCount = acc_get_num_devices(acc_device_nvidia);
		if (devId < 0 || devId >= devCount) {
			printf("The specified device ID is not supported.\n");
			exit(1);
		}
	}
	if (devId != -1) {
		acc_set_device_num(devId, acc_device_nvidia);
	}
	// creates a context on the GPU just to
	// exclude initialization time from computations
	acc_init(acc_device_nvidia);

	// print device id
	devId = acc_get_device_num(acc_device_nvidia);
	printf("Running on GPU with ID %d.\n\n", devId);

}
#endif

/*
* setting values, init mpi, omp etc
*/
static void Init(struct JacobiData *data)
{
	/* default medium */
	data->iCols      = 2000;
	data->iRows      = 2000;
	data->fAlpha     = 0.8f;
	data->fRelax     = 1.0f;
	data->fTolerance = 1e-10f;
	data->iIterMax   = 50;
#ifdef READ_INPUT
	printf("Input n - matrix size in x direction:                 ");
	scanf("%d", &data->iCols);
	printf("\nInput m - matrix size in y direction:               ");
	scanf("%d", &data->iRows);
	printf("\nInput alpha - Helmholtz constant:                   ");
	scanf("%f", &data->fAlpha);
	printf("\nInput relax - Successive over-relaxation parameter: ");
	scanf("%f", &data->fRelax);
	printf("\nInput tol - error tolerance for iterrative solver:  ");
	scanf("%f", &data->fTolerance);
	printf("\nInput mits - Maximum iterations for solver:         ");
	scanf("%d", &data->iIterMax);
#elif defined DATA_LARGE
	data->iCols      = 7000;
	data->iRows      = 7000;
	data->fAlpha     = 0.8f;
	data->fRelax     = 1.0f;
	data->fTolerance = 1e-12f;
	data->iIterMax   = 2;

#elif defined DATA_SMALL
	data->iCols      = 200;
	data->iRows      = 200;
	data->fAlpha     = 0.8f;
	data->fRelax     = 1.0f;
	data->fTolerance = 1e-7f;
	data->iIterMax   = 1000;
#endif
	printf("\n-> matrix size: %dx%d"
	"\n-> alpha: %f"
	"\n-> relax: %f"
	"\n-> tolerance: %f"
	"\n-> #of iterations: %d \n\n",
	data->iCols, data->iRows, data->fAlpha, data->fRelax,
	data->fTolerance, data->iIterMax);


	/* MPI values, set to defaults to avoid data inconsistency */
	data->iMyRank   = 0;
	data->iNumProcs = 1;
	data->iRowFirst = 0;
	data->iRowLast  = data->iRows - 1;

	/* memory allocation for serial & omp */
	data->afU = (float*) malloc(data->iRows * data->iCols * sizeof(float));
	data->afF = (float*) malloc(data->iRows * data->iCols * sizeof(float));

	/* calculate dx and dy */
	data->fDx = 2.0f / (data->iCols - 1);
	data->fDy = 2.0f / (data->iRows - 1);

	data->iIterCount = 0;

	return;
}

/*
* final cleanup routines
*/
void Finish(struct JacobiData * data)
{
	free (data->afU);
	free (data->afF);

	return;
}

/*
* print result summary
*/
static void PrintResults(const struct JacobiData * data)
{
	if (data->iMyRank == 0)
	{
		printf(" Number of iterations : %d\n", data->iIterCount);
		printf(" Residual             : %le\n", data->fResidual);
		printf(" Solution Error       : %1.12lf\n", data->fError);
		printf(" Elapsed Time         : %5.7lf\n", 
		data->fTimeStop-data->fTimeStart);
		printf(" MFlops               : %6.6lf\n", 
		0.000013 * data->iIterCount * (data->iCols - 2) * (data->iRows - 2)
		/ (data->fTimeStop - data->fTimeStart));
	}
	return;
}

/*
* Initializes matrix
* Assumes exact solution is u(x,y) = (1-x^2)*(1-y^2)
*/
static void InitializeMatrix(struct JacobiData * data)
{
	int i, j, xx, yy, xx2, yy2;
	//    int i, j;
	//    float xx, yy, xx2, yy2;
	
	/* Initialize initial condition and RHS */
	for (j = data->iRowFirst; j <= data->iRowLast; j++)
	{
		for (i = 0; i < data->iCols; i++)
		{
			/* TODO: check if this values have to be ints or doubles */
			xx = (int) (-1.0f + data->fDx * i);
			yy = (int) (-1.0f + data->fDy * j);
			//            xx =  (-1.0f + data->fDx * i);
			//            yy =  (-1.0f + data->fDy * j);

			xx2 = xx * xx;
			yy2 = yy * yy;

			U(j,i) = 0.0;
			F(j,i) = -data->fAlpha * (1.0f - xx2) * (1.0f - yy2)
			+ 2.0f * (-2.0f + xx2 + yy2);
		}
	}
}

/*
* Checks error between numerical and exact solution
*/
static void CheckError(struct JacobiData * data)
{
	float error = 0.0f;
	int i, j;
	float xx, yy, temp;

	for (j = data->iRowFirst; j <= data->iRowLast; j++)
	{
		if ((data->iMyRank != 0 && j == data->iRowFirst) || 
				(data->iMyRank != data->iNumProcs - 1 && j == data->iRowLast))
		continue;

		for (i = 0; i < data->iCols; i++)
		{
			xx   = -1.0f + data->fDx * i;
			yy   = -1.0f + data->fDy * j;
			temp = U(j,i) - (1.0f - xx * xx) * (1.0f - yy * yy);
			error += temp * temp;
		}
	}

	data->fError = sqrtf(error) / (data->iCols * data->iRows);
	
	return;
}


int main (int argc, char** argv)
{
	int retVal = 0;    /* return value */

	struct JacobiData myData;


	/* sets default values or reads from stdin
	* inits MPI and OpenMP if needed
	* distribute MPI data, calculate MPI bounds
	*/
	Init(&myData);

#ifdef _ACCEL
	InitGPU(argc, argv);
#endif

	if (myData.afU && myData.afF)
	{
		/* matrix init */
		InitializeMatrix(&myData);

		/* starting timer */
		myData.fTimeStart = GetRealTime();

		/* running calculations */
		Jacobi(myData.iRows, myData.iCols, myData.iRowLast, myData.iIterMax, myData.fAlpha, myData.fRelax, myData.fTolerance, myData.fDx, myData.fDy, myData.afU, myData.afF, &(myData.fResidual), &myData.iIterCount);

		/* stopping timer */
		myData.fTimeStop = GetRealTime();

		/* error checking */
		CheckError(&myData);

		/* print result summary */
		PrintResults(&myData);
	}
	else
	{
		printf(" Memory allocation failed ...\n");
		retVal = -1;
	}

	/* cleanup */
	Finish(&myData);
	
	return retVal;
}
