#define THREADSPERBLOCK 256

module JacobiMod
  use VariableDef
  implicit none 

contains

  subroutine Jacobi(iRows, iCols, iIterMax, fAlpha, fRelax, fTolerance, fDx, afU, afF, fResidual, iIterCount)
    implicit none
    !********************************************************************
    ! Subroutine Jacobi                                                 *
    ! Solves poisson equation on rectangular grid assuming :            *
    ! (1) Uniform discretization in each direction, and                 *
    ! (2) Dirichlect boundary conditions                                *
    ! Jacobi method is used in this routine                             *
    !                                                                   *
    ! Input : n,m   Number of grid points in the X/Y directions         *
    !         dx,dy Grid spacing in the X/Y directions                  *
    !         alpha Helmholtz eqn. coefficient                          *
    !         omega Relaxation factor                                   *
    !         myData%afF(n,m) Right hand side function                  *
    !         myData%afU(n,m) Dependent variable/Solution               *
    !         tol    Tolerance for iterative solver                     *
    !         maxit  Maximum number of iterations                       *
    !                                                                   *
    ! Output : myData%afU(n,m) - Solution                               *
    !******************************************************************** 

    !.. Formal Arguments .. 
    integer, intent(in) :: iRows
    integer, intent(in) :: iCols
    integer, intent(in) :: iIterMax
    real, intent(in) :: fAlpha
    real, intent(in) :: fRelax
    real, intent(in) :: fTolerance
    real, intent(in) :: fDx
    real, intent(inout) :: afU(0:iCols-1, 0:iRows-1)
    real, intent(in) :: afF(0:iCols-1, 0:iRows-1)
    real, intent(inout) :: fResidual
    integer, intent(inout) :: iIterCount

    !.. Local Scalars .. 
    integer :: i, j, iErr
    real :: ax, ay, b, residual, fLRes, tmpResd

    !.. Local Arrays .. 
    real, allocatable :: uold(:,:)

    allocate(uold (0 : iCols -1, 0 : iRows -1))

    ! ... Executable Statements ...
    ! Initialize coefficients 

    if (allocated(uold)) then    
       ax = 1.0 / (fDx * fDx)      ! X-direction coef 
       ay = 1.0 / (fDx * fDx)      ! Y-direction coef
       b = -2.0 * (ax + ay) - fAlpha      ! Central coeff  
       residual = 10.0 * fTolerance

       ! Moves data only once        
!$acc data region local(uold) copyin(afF) copy(afU)
       do while (iIterCount < iIterMax .and. residual > fTolerance)
          residual = 0.0

          ! Creates an accelerator region for the swapping (copy needed data)
!$acc region        

          ! TODO: Modify the loop schedule to distribute outer loop to the blocks
          ! and inner loop to threads within the block
!$acc do parallel
          ! Copy new solution into old
          do j = 0, iRows-1
!$acc do vector(THREADSPERBLOCK)
             do i = 0, iCols-1
                uold(i,j) = afU(i,j)
             enddo
          enddo
!$acc end region

          ! Creates an accelerator region for the computations (copy needed data)
!$acc region
          ! TODO: Modify the loop schedule to distribute outer loop to the blocks
          ! and inner loop to threads within the block
!$acc do parallel
          ! Compute stencil, residual, & update
          do j = 1, iRows - 2
!$acc do vector(THREADSPERBLOCK)
             do i = 1, iCols - 2
                ! Evaluate residual 
                fLRes = (ax * (uold(i-1, j) + uold(i+1, j)) &
                     + ay * (uold(i, j-1) + uold(i, j+1)) &
                     + b * uold(i, j) - afF(i, j)) / b

                ! Update solution 
                afU(i, j) = uold(i, j) - fRelax * fLRes

                ! Accumulate residual error
                residual = residual + fLRes * fLRes
             end do
          end do
!$acc end region        

          ! Error check 
          iIterCount = iIterCount + 1      
          residual = SQRT(residual) / REAL(iCols * iRows)

          ! End iteration loop 
       end do
!$acc end data region

       fResidual = residual
       deallocate(uold)
    else
       write (*,*) 'Error: cant allocate memory'
       stop
    end if
  end subroutine Jacobi

end module JacobiMod
