/* GPGPU computation of jacobi algorithm with SINGLE PRECISION!
* 
******************************************************************
* Subroutine Jacobi
* Solves poisson equation on rectangular grid assuming : 
* (1) Uniform discretization in each direction, and 
* (2) Dirichlect boundary conditions 
* 
* Jacobi method is used in this routine 
*
* Input : n,m   Number of grid points in the X/Y directions 
*         dx,dy Grid spacing in the X/Y directions 
*         alpha Helmholtz eqn. coefficient 
*         omega Relaxation factor 
*         f(n,m) Right hand side function 
*         u(n,m) Dependent variable/Solution
*         tolerance Tolerance for iterative solver 
*         maxit  Maximum number of iterations 
*
* Output : u(n,m) - Solution 
*****************************************************************
*/
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "jacobi.h"

#define THREADSPERBLOCK 256

void Jacobi (int iRows, int iCols, int iIterMax, float fAlpha, float fRelax, float fTolerance, float fDx, float fDy, float *afU, float *afF, float* fResidual, int* iIterCount, cl_command_queue oclCommandQueue, cl_context oclContext, cl_kernel oclKernel)
{
	cl_int err;
	
	float ax, ay, b, residual = 0.0f;
	int n = iCols * iRows;
	
	float *h_resArray = (float*)malloc(iRows * sizeof(float)); // array to store interim residuals
	cl_mem d_uold, d_u, d_f, d_resArray; // memory on device

	// TODO: allocate memory on the device for u, uold and f
	// TODO: Copy data from host to device
	d_uold = clCreateBuffer(oclContext, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR, n * sizeof(float), afU, &err);
	checkErr(err,__FILE__,__LINE__);
	d_u = clCreateBuffer(oclContext, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR, n * sizeof(float), afU, &err);
	checkErr(err,__FILE__,__LINE__);
	d_f =  clCreateBuffer(oclContext, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, n * sizeof(float), afF, &err);
	checkErr(err,__FILE__,__LINE__);

	// TODO: allocate memory on the device to store iterim residuals
	d_resArray = clCreateBuffer(oclContext, CL_MEM_WRITE_ONLY, iRows * sizeof(float), 0, &err);
	checkErr(err,__FILE__,__LINE__);

	ax = 1.0f / (fDx * fDx);      // X-direction coef
	ay = 1.0f / (fDy * fDy);      // Y_direction coef
	b = -2.0f * (ax + ay) - fAlpha;     // Central coeff
	residual = 10.0f * fTolerance;

	while (*iIterCount < iIterMax && residual > fTolerance) 
	{
		residual = 0.0f; 
		
		// swap u and uold
		cl_mem utmp;
		utmp = d_uold;
		d_uold = d_u;
		d_u = utmp;

		// TODO: Compute the total number of threads (hint: _at least_ a number of 'iRows')
		size_t threads_per_block[1], total_num_threads[1];
		threads_per_block[0] = THREADSPERBLOCK;
		if (iRows%THREADSPERBLOCK == 0) {
			total_num_threads[0] = iRows;
		} else {
			total_num_threads[0] = THREADSPERBLOCK * (iRows/THREADSPERBLOCK + 1);
		}
		
		// TODO: Set kernel arguments
		int i = 0;
		err = clSetKernelArg(oclKernel, i++, sizeof(cl_mem), &d_u); checkErr(err,__FILE__,__LINE__);
		err = clSetKernelArg(oclKernel, i++, sizeof(cl_mem), &d_uold); checkErr(err,__FILE__,__LINE__);
		err = clSetKernelArg(oclKernel, i++, sizeof(cl_mem), &d_f); checkErr(err,__FILE__,__LINE__);
		err = clSetKernelArg(oclKernel, i++, sizeof(cl_int), &iRows); checkErr(err,__FILE__,__LINE__);
		err = clSetKernelArg(oclKernel, i++, sizeof(cl_int), &iCols); checkErr(err,__FILE__,__LINE__);
		err = clSetKernelArg(oclKernel, i++, sizeof(cl_float), &ax); checkErr(err,__FILE__,__LINE__);
		err = clSetKernelArg(oclKernel, i++, sizeof(cl_float), &ay); checkErr(err,__FILE__,__LINE__);
		err = clSetKernelArg(oclKernel, i++, sizeof(cl_float), &b); checkErr(err,__FILE__,__LINE__);
		err = clSetKernelArg(oclKernel, i++, sizeof(cl_float), &fRelax); checkErr(err,__FILE__,__LINE__);
		err = clSetKernelArg(oclKernel, i++, sizeof(cl_mem), &d_resArray); checkErr(err,__FILE__,__LINE__);

		// TODO: Launch the GPU kernel
		err = clEnqueueNDRangeKernel(oclCommandQueue, oclKernel, 1, NULL, total_num_threads, threads_per_block, 0, NULL, NULL);
		checkErr(err,__FILE__,__LINE__);

		// TODO: Copy the iterim residuals from the device to the host to finally reduce them on the host
		err = clEnqueueReadBuffer(oclCommandQueue, d_resArray, CL_TRUE, 0, iRows * sizeof(float), h_resArray, 0, NULL, NULL);
		checkErr(err,__FILE__,__LINE__);

		residual = reduceOnCPU(h_resArray, iRows);

		(*iIterCount)++;
		residual = sqrtf(residual) / (iCols * iRows);
	} // while

	*fResidual = residual;
	
	// TODO: Copy final solution u from device to host
	err = clEnqueueReadBuffer(oclCommandQueue, d_u, CL_TRUE, 0, n * sizeof(float), afU, 0 ,NULL, NULL);

	// TODO: Free all allocated memory on the device
	clReleaseMemObject(d_uold);
	clReleaseMemObject(d_u);
	clReleaseMemObject(d_f);
	clReleaseMemObject(d_resArray);

	free(h_resArray);
}

float reduceOnCPU(float *array, int size) {
	float sum = 0.0f;
	int i;
	for (i=0; i<size; i++) {
		sum+=array[i];
	}
	return sum;
}

void checkErr(cl_int err, const char* file, const int line)
{
	char * errMessage;
	switch (err)
	{
	case CL_SUCCESS:                                    //  0
		errMessage= "CL_SUCCESS"; return;
	case CL_DEVICE_NOT_FOUND:                           //  -1
		errMessage= "CL_DEVICE_NOT_FOUND"; break;
	case CL_DEVICE_NOT_AVAILABLE:                       //  -2
		errMessage= "CL_DEVICE_NOT_AVAILABLE"; break;
	case CL_COMPILER_NOT_AVAILABLE:                     //  -3
		errMessage= "CL_COMPILER_NOT_AVAILABLE"; break;
	case CL_MEM_OBJECT_ALLOCATION_FAILURE:              //  -4
		errMessage= "CL_MEM_OBJECT_ALLOCATION_FAILURE"; break;
	case CL_OUT_OF_RESOURCES:                           //  -5
		errMessage= "CL_OUT_OF_RESOURCES"; break;
	case CL_OUT_OF_HOST_MEMORY:                         //  -6
		errMessage= "CL_OUT_OF_HOST_MEMORY"; break;
	case CL_PROFILING_INFO_NOT_AVAILABLE:               //  -7
		errMessage= "CL_PROFILING_INFO_NOT_AVAILABLE"; break;
	case CL_MEM_COPY_OVERLAP:                           //  -8
		errMessage= "CL_MEM_COPY_OVERLAP"; break;
	case CL_IMAGE_FORMAT_MISMATCH:                      //  -9
		errMessage= "CL_IMAGE_FORMAT_MISMATCH"; break;
	case CL_IMAGE_FORMAT_NOT_SUPPORTED:                 //  -10
		errMessage= "CL_IMAGE_FORMAT_NOT_SUPPORTED"; break;
	case CL_BUILD_PROGRAM_FAILURE:                      //  -11
		errMessage= "CL_BUILD_PROGRAM_FAILURE"; break;
	case CL_MAP_FAILURE:                                //  -12
		errMessage= "CL_MAP_FAILURE"; break;
	case CL_MISALIGNED_SUB_BUFFER_OFFSET:               //  -13
		errMessage= "CL_MISALIGNED_SUB_BUFFER_OFFSET"; break;
	case CL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST:  //  -14
		errMessage= "CL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST"; break;
	case CL_INVALID_VALUE:                              //  -30
		errMessage= "CL_INVALID_VALUE"; break;
	case CL_INVALID_DEVICE_TYPE:                        //  -31
		errMessage= "CL_INVALID_DEVICE_TYPE"; break;
	case CL_INVALID_PLATFORM:                           //  -32
		errMessage= "CL_INVALID_PLATFORM"; break;
	case CL_INVALID_DEVICE:                             //  -33
		errMessage= "CL_INVALID_DEVICE"; break;
	case CL_INVALID_CONTEXT:                            //  -34
		errMessage= "CL_INVALID_CONTEXT"; break;
	case CL_INVALID_QUEUE_PROPERTIES:                   //  -35
		errMessage= "CL_INVALID_QUEUE_PROPERTIES"; break;
	case CL_INVALID_COMMAND_QUEUE:                      //  -36
		errMessage= "CL_INVALID_COMMAND_QUEUE"; break;
	case CL_INVALID_HOST_PTR:                           //  -37
		errMessage= "CL_INVALID_HOST_PTR"; break;
	case CL_INVALID_MEM_OBJECT:                         //  -38
		errMessage= "CL_INVALID_MEM_OBJECT"; break;
	case CL_INVALID_IMAGE_FORMAT_DESCRIPTOR:            //  -39
		errMessage= "CL_INVALID_IMAGE_FORMAT_DESCRIPTOR"; break;
	case CL_INVALID_IMAGE_SIZE:                         //  -40
		errMessage= "CL_INVALID_IMAGE_SIZE"; break;
	case CL_INVALID_SAMPLER:                            //  -41
		errMessage= "CL_INVALID_SAMPLER"; break;
	case CL_INVALID_BINARY:                             //  -42
		errMessage= "CL_INVALID_BINARY"; break;
	case CL_INVALID_BUILD_OPTIONS:                      //  -43
		errMessage= "CL_INVALID_BUILD_OPTIONS"; break;
	case CL_INVALID_PROGRAM:                            //  -44
		errMessage= "CL_INVALID_PROGRAM"; break;
	case CL_INVALID_PROGRAM_EXECUTABLE:                 //  -45
		errMessage= "CL_INVALID_PROGRAM_EXECUTABLE"; break;
	case CL_INVALID_KERNEL_NAME:                        //  -46
		errMessage= "CL_INVALID_KERNEL_NAME"; break;
	case CL_INVALID_KERNEL_DEFINITION:                  //  -47
		errMessage= "CL_INVALID_KERNEL_DEFINITION"; break;
	case CL_INVALID_KERNEL:                             //  -48
		errMessage= "CL_INVALID_KERNEL"; break;
	case CL_INVALID_ARG_INDEX:                          //  -49
		errMessage= "CL_INVALID_ARG_INDEX"; break;
	case CL_INVALID_ARG_VALUE:                          //  -50
		errMessage= "CL_INVALID_ARG_VALUE"; break;
	case CL_INVALID_ARG_SIZE:                           //  -51
		errMessage= "CL_INVALID_ARG_SIZE"; break;
	case CL_INVALID_KERNEL_ARGS:                        //  -52
		errMessage= "CL_INVALID_KERNEL_ARGS"; break;
	case CL_INVALID_WORK_DIMENSION:                     //  -53
		errMessage= "CL_INVALID_WORK_DIMENSION"; break;
	case CL_INVALID_WORK_GROUP_SIZE:                    //  -54
		errMessage= "CL_INVALID_WORK_GROUP_SIZE"; break;
	case CL_INVALID_WORK_ITEM_SIZE:                     //  -55
		errMessage= "CL_INVALID_WORK_ITEM_SIZE"; break;
	case CL_INVALID_GLOBAL_OFFSET:                      //  -56
		errMessage= "CL_INVALID_GLOBAL_OFFSET"; break;
	case CL_INVALID_EVENT_WAIT_LIST:                    //  -57
		errMessage= "CL_INVALID_EVENT_WAIT_LIST"; break;
	case CL_INVALID_EVENT:                              //  -58
		errMessage= "CL_INVALID_EVENT"; break;
	case CL_INVALID_OPERATION:                          //  -59
		errMessage= "CL_INVALID_OPERATION"; break;
	case CL_INVALID_GL_OBJECT:                          //  -60
		errMessage= "CL_INVALID_GL_OBJECT"; break;
	case CL_INVALID_BUFFER_SIZE:                        //  -61
		errMessage= "CL_INVALID_BUFFER_SIZE"; break;
	case CL_INVALID_MIP_LEVEL:                          //  -62
		errMessage= "CL_INVALID_MIP_LEVEL"; break;
	case CL_INVALID_GLOBAL_WORK_SIZE:                   //  -63
		errMessage= "CL_INVALID_GLOBAL_WORK_SIZE"; break;
		//case CL_INVALID_PROPERTY:                           //  -64
		//    errMessage= "CL_INVALID_PROPERTY"; break;
	default:
		errMessage= "AN UNKNOWN OpenCL ERROR CODE"; break;
	}
	fprintf(stderr, "%s: OpenCL error in line %d: %s\n", file, line, errMessage);
	exit(EXIT_FAILURE);
}
