/* GPGPU computation of jacobi algorithm with SINGLE PRECISION!
* 
******************************************************************
* Subroutine Jacobi
* Solves poisson equation on rectangular grid assuming : 
* (1) Uniform discretization in each direction, and 
* (2) Dirichlect boundary conditions 
* 
* Jacobi method is used in this routine 
*
* Input : n,m   Number of grid points in the X/Y directions 
*         dx,dy Grid spacing in the X/Y directions 
*         alpha Helmholtz eqn. coefficient 
*         omega Relaxation factor 
*         f(n,m) Right hand side function 
*         u(n,m) Dependent variable/Solution
*         tolerance Tolerance for iterative solver 
*         maxit  Maximum number of iterations 
*
* Output : u(n,m) - Solution 
*****************************************************************
*/
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "jacobi.h"

#define THREADSPERBLOCK 256

#define U(j,i) afU[(j) * iCols + (i)]
#define F(j,i) afF[(j) * iCols + (i)]
#define UOLD(j,i) uold[(j) * iCols + (i)]

void Jacobi (int iRows, int iCols, int iRowLast, int iIterMax, float fAlpha, float fRelax, float fTolerance, float fDx, float fDy, float *restrict afU, float *restrict afF, float* restrict fResidual, int* restrict iIterCount)
{
	int i, j;   
	float ax, ay, b, fLRes, residual = 0.0f;
	int n = iCols * iRows;
	
	float *restrict uold = (float*)malloc(iCols * iRows * sizeof(float));

	if (uold)
	{
		ax = 1.0f / (fDx * fDx);      // X-direction coef
		ay = 1.0f / (fDy * fDy);      // Y_direction coef
		b = -2.0f * (ax + ay) - fAlpha;     // Central coef
		residual = 10.0f * fTolerance;


		// Moves data only once 
#pragma acc data region local(uold[0:n-1]) copyin(afF[0:n-1]) copy(afU[0:n-1])
		{
			while (*iIterCount < iIterMax && residual > fTolerance) 
			{
				residual = 0.0f; 
				
				// Creates an accelerator region for the swapping (copy needed data)
#pragma acc region
				{
					// TODO: Modify the loop schedule to distribute outer loop to the blocks
					// and inner loop to threads within the block
#pragma acc for parallel
					// copy new solution into old 
					for (j = 0; j < iRows; j++)
					{
#pragma acc for vector(THREADSPERBLOCK)
						for (i = 0; i < iCols; i++)
						{
							UOLD(j,i) = U(j,i);
						}
					}
				}  // acc region


				// Creates an accelerator region for the computations (copy needed data)
#pragma acc region
				{
					// TODO Modify the loop schedule to distribute outer loop to blocks
					// and inner loop to threds within the block
#pragma acc for parallel
					// compute stencil, residual and update
					for (j = 1; j <= iRowLast - 1; j++)
					{
#pragma acc for vector(THREADSPERBLOCK)
						for (i = 1; i <= iCols - 2; i++)
						{
							fLRes = ( ax * (UOLD(j, i-1) + UOLD(j, i+1))
							+ ay * (UOLD(j-1, i) + UOLD(j+1, i))
							+  b * UOLD(j, i) - F(j, i)) / b;
							U(j,i) = UOLD(j,i) - fRelax * fLRes;
							residual += fLRes * fLRes;
						}
					}
				} // acc region

				(*iIterCount)++;
				residual = sqrtf(residual) / (iCols * iRows);
			} // while
		} // data acc
		*fResidual = residual;
		free(uold);
	}
	else 
	{
		fprintf(stderr,"Error: can't allocate memory\n");
		exit(1);
	}
}

