/* GPGPU computation of jacobi algorithm with SINGLE PRECISION!
* 
******************************************************************
* Subroutine Jacobi
* Solves poisson equation on rectangular grid assuming : 
* (1) Uniform discretization in each direction, and 
* (2) Dirichlect boundary conditions 
* 
* Jacobi method is used in this routine 
*
* Input : n,m   Number of grid points in the X/Y directions 
*         dx,dy Grid spacing in the X/Y directions 
*         alpha Helmholtz eqn. coefficient 
*         omega Relaxation factor 
*         f(n,m) Right hand side function 
*         u(n,m) Dependent variable/Solution
*         tolerance Tolerance for iterative solver 
*         maxit  Maximum number of iterations 
*
* Output : u(n,m) - Solution 
*****************************************************************
*/
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "jacobi.cuh"

#define THREADSPERBLOCK 256

#define U(j,i) u[(j) * cols + (i)]
#define F(j,i) f[(j) * cols + (i)]
#define UOLD(j,i) uold[(j) * cols + (i)]


/* Performs a single Jacobi iteration on the GPU
* --------------------------------------------------------
* Outer loop has to be distributed among the threads
* Inner loop shall be executed sequentially by each thread
*/

// TODO: Define this method as GPU kernel
__global__
void oneJacobiIteration(float* u, float* uold, float* f, int rows, int cols, float ax, float ay, float b, float fRelax, float* residuals) { 

	// TODO: Compute the global index within the arrays (with respect to block and thread ids)
	int idx = blockIdx.x * blockDim.x + threadIdx.x;

	// initialize interim array of residuals
	// check that you do not access memory that has not been allocated
	if (idx < rows) { residuals[idx] = 0.0f; } // ERROR was here

	// TODO: Express the outer loop by starting the kernel with 
	// a corrsponding number of threads

	// TODO: Verify that it is not operated on the boundary elements
	if (idx >= 1 && idx <= rows - 2) {
		for (int i = 1; i <= cols - 2; i++)
		{
			float fLRes = ( ax * (UOLD(idx, i-1) + UOLD(idx, i+1))
			+ ay * (UOLD(idx-1, i) + UOLD(idx+1, i))
			+  b * UOLD(idx, i) - F(idx, i)) / b;
			U(idx,i) = UOLD(idx,i) - fRelax * fLRes;
			residuals[idx] += fLRes * fLRes;
		}
	}
}


void Jacobi (int iRows, int iCols, int iIterMax, float fAlpha, float fRelax, float fTolerance, float fDx, float fDy, float *afU, float *afF, float* fResidual, int* iIterCount)
{
	cudaError_t err;
	float ax, ay, b, residual = 0.0f;
	int n = iCols * iRows;
	
	float *h_resArray = (float*)malloc(iRows * sizeof(float)); // array to store interim residuals
	float *d_uold, *d_u, *d_f, *d_resArray; // pointer to device

	// TODO: allocate memory on the device for u, uold and f
	err = cudaMalloc((void**)&d_uold, n * sizeof(float));
	checkErr(err, __FILE__, __LINE__);
	err = cudaMalloc((void**)&d_u, n * sizeof(float));
	checkErr(err, __FILE__, __LINE__);
	err = cudaMalloc((void**)&d_f, n * sizeof(float));
	checkErr(err, __FILE__, __LINE__);

	// TODO: allocate memory on the device to store iterim residuals
	err = cudaMalloc((void**)&d_resArray, iRows * sizeof(float));
	checkErr(err, __FILE__, __LINE__);

	ax = 1.0f / (fDx * fDx);      // X-direction coef
	ay = 1.0f / (fDy * fDy);      // Y_direction coef
	b = -2.0f * (ax + ay) - fAlpha;     // Central coeff
	residual = 10.0f * fTolerance;

	// TODO: copy data from host to device
	err = cudaMemcpy(d_u, afU, n*sizeof(float), cudaMemcpyHostToDevice);
	checkErr(err, __FILE__, __LINE__);
	err = cudaMemcpy(d_uold, afU, n*sizeof(float), cudaMemcpyHostToDevice);
	checkErr(err, __FILE__, __LINE__);
	err = cudaMemcpy(d_f, afF, n*sizeof(float), cudaMemcpyHostToDevice);
	checkErr(err, __FILE__, __LINE__);

	while (*iIterCount < iIterMax && residual > fTolerance) 
	{
		residual = 0.0f; 
		
		// swap u and uold
		float *utmp;
		utmp = d_uold;
		d_uold = d_u;
		d_u = utmp;

		dim3 threads_per_block(THREADSPERBLOCK);
		dim3 blocks_per_grid;

		// TODO: Compute the number of blocks_per_grid
		// The TOTAL number of threads shall be at least 'iRows'
		if (iRows%THREADSPERBLOCK == 0) {
			blocks_per_grid = dim3(iRows/THREADSPERBLOCK);
		} else {
			blocks_per_grid = dim3(iRows/THREADSPERBLOCK+1);
		}

		// TODO: Launch the GPU kernel
		oneJacobiIteration<<<blocks_per_grid,threads_per_block>>>(d_u, d_uold, d_f, iRows, iCols, ax, ay, b, fRelax, d_resArray);

		// TODO: Copy the iterim residuals from the device to the host to finally reduce them on the host
		err = cudaMemcpy(h_resArray, d_resArray, iRows*sizeof(float), cudaMemcpyDeviceToHost);
		checkErr(err, __FILE__, __LINE__);

		residual = reduceOnCPU(h_resArray, iRows);

		(*iIterCount)++;
		residual = sqrtf(residual) / (iCols * iRows);
	} // while

	*fResidual = residual;

	// TODO: Copy final solution u from device to host
	err = cudaMemcpy(afU, d_u, n*sizeof(float), cudaMemcpyDeviceToHost);
	checkErr(err, __FILE__, __LINE__);

	// TODO: Free all allocated memory on the device
	cudaFree(d_uold);
	cudaFree(d_u);
	cudaFree(d_f);
	cudaFree(d_resArray);

	free(h_resArray);
}

float reduceOnCPU(float *array, int size) {
	int i;
	float sum = 0.0f;
	for (i=0; i<size; i++) {
		sum+=array[i];
	}
	return sum;
}

// Checks whether a CUDA error occured
// If so, the error message is printed and the program exits
inline void checkErr(cudaError_t err, const char* file, const int line)
{
	if(cudaSuccess != err)
	{
		fprintf(stderr, "%s: Cuda error in line %d: %s.\n", file, line, cudaGetErrorString(err) );
		exit(-1);
	}
}
