program MAIN
  !***********************************************************************
  ! program to solve a finite difference                                 * 
  ! discretization of Helmholtz equation :                               *
  ! (d2/dx2)u + (d2/dy2)u - alpha u = f                                  *
  ! using Jacobi iterative method.                                       *
  !                                                                      *
  ! Modified: Abdelali Malih,    Aachen University (RWTH), 2007          *
  ! Modified: Sanjiv Shah,       Kuck and Associates, Inc. (KAI), 1998   *
  ! Author  : Joseph Robicheaux, Kuck and Associates, Inc. (KAI), 1998   *
  !                                                                      * 
  ! Directives are used in this code to achieve paralleism.              *
  ! All do loops are parallized with default 'static' scheduling.        *
  !                                                                      * 
  ! Input :  n - grid dimension in x direction                           *
  !          m - grid dimension in y direction                           *
  !          alpha - Helmholtz constant (always greater than 0.0)        *
  !          tol   - error tolerance for iterative solver                *
  !          relax - Successice over relaxation parameter                *
  !          mits  - Maximum iterations for iterative solver             *
  !                                                                      *
  ! On output                                                            *
  !       : u(n,m) - Dependent variable (solutions)                      *
  !       : f(n,m) - Right hand side function                            *
  !***********************************************************************

  use VariableDef
  use JacobiMod
  use RealTime
#ifdef _ACCEL
  use accel_lib
#endif
  implicit none

  TYPE(JacobiData) :: myData


  !   sets default values or reads from stdin
  !    * inits MPI and OpenMP if needed
  !    * distribute MPI data, calculate MPI bounds
  !    */
  call Init(mydata)

#ifdef _ACCEL
  call InitGPU()
#endif


  if ( allocated(myData%afU) .and. allocated(myData%afF) ) then
     !        /* matrix init */

     call InitializeMatrix(myData)

     !        /* starting timer */
     mydata%fTimeStart = GetRealTime()

     !        /* running calculations */
     call Jacobi(myData%iRows, myData%iCols, myData%iIterMax, myData%fAlpha, myData%fRelax, &
          myData%fTolerance, myData%fDx, myData%afU, myData%afF, myData%fResidual, myData%iIterCount)

     !        /* stopping timer */
     mydata%fTimeStop = GetRealTime()

     !        /* error checking */
     call CheckError(myData)

     !        /* print result summary */
     call PrintResults(myData)


  else
     write (*,*) " Memory allocation failed ...\n"
  end if

  !    /* cleanup */
  call Finish(myData)

end program MAIN

#ifdef _ACCEL
subroutine InitGPU()
#ifdef _ACCEL
  use accel_lib
#endif
  implicit none
  character :: argv*10
  integer :: iargc, argc, devId, devCount

  argc = iargc()

  ! gets the device id (if specified) to run on
  devId = -1
  if (argc > 1) then
     call getarg(1,devId);
     devCount = acc_get_num_devices(acc_device_nvidia);
     if (devId < 0 .OR. devId >= devCount) then
        write(*,*) "The specified device ID is not supported."
        stop
     end if
  end if
  if (devId /= -1) then
     call acc_set_device_num(devId, acc_device_nvidia)
  end if
  ! creates a context on the GPU just to
  ! exclude initialization time from computations
  call acc_init(acc_device_nvidia)

  ! print device id
  !devId = acc_get_device_num(acc_device_nvidia)

end subroutine InitGPU
#endif

subroutine Init (myData)
  use VariableDef
  implicit none
  type(JacobiData), intent(inout) :: myData
  integer :: iErr, i
  ! default medium
  myData%iCols      = 5000
  myData%iRows      = 5000
  myData%fAlpha     = 0.8
  myData%fRelax     = 1.0
  myData%fTolerance = 1e-12
  myData%iIterMax   = 20
#ifdef READ_INPUT
  write (*,*) 'Input n - matrix size in x direction: '
  read (5,*) myData%iCols
  write (*,*) 'Input m - matrix size in y direction: '
  read (5,*) myData%iRows
  write (*,*) 'Input alpha - Helmholts constant:'
  read (5,*) myData%fAlpha
  write (*,*) 'Input relax - Successive over-relaxation parameter:'
  read (5,*) myData%fRelax
  write (*,*) 'Input tol - error tolerance for iterative solver:'
  read (5,*) myData%fTolerance
  write (*,*) 'Input mits - Maximum iterations for solver:'
  read (5,*) myData%iIterMax
#elif defined DATA_LARGE
  myData%iCols      = 7000
  myData%iRows      = 7000
  myData%fAlpha     = 0.8
  myData%fRelax     = 1.0
  myData%fTolerance = 1e-12
  myData%iIterMax   = 2

#elif defined DATA_SMALL
  myData%iCols      = 200
  myData%iRows      = 200
  myData%fAlpha     = 0.8
  myData%fRelax     = 1.0
  myData%fTolerance = 1e-7
  myData%iIterMax   = 1000
#endif
  write (*,327) "-> matrix size:", myData%iCols, myData%iRows
  write (*,329) "-> alpha: " , myData%fAlpha
  write (*,329) "-> relax: ", myData%fRelax
  write (*,329) "-> tolerance: ", myData%fTolerance
  write (*,328) "-> #of iterations: ", myData%iIterMax
327 format (A22, I10, ' x ', I10)
328 format (A22, I10)
329 format (A22, F10.6)

  ! memory allocation for serial & omp
  allocate(myData%afU (0 : myData%iCols -1, 0 : myData%iRows -1))
  allocate(myData%afF (0 : myData%iCols -1, 0 : myData%iRows -1))

  ! calculate dx and dy
  myData%fDx = 2.0 / REAL(myData%iCols - 1)
  myData%fDy = 2.0 / REAL(myData%iRows - 1)

  myData%iIterCount = 0

end subroutine Init

subroutine InitializeMatrix (myData)
  !*********************************************************************
  ! Initializes data                                                   *
  ! Assumes exact solution is u(x,y) = (1-x^2)*(1-y^2)                 *
  !                                                                    *
  !*********************************************************************
  use VariableDef
  implicit none

  type(JacobiData), intent(inout) :: myData 
  !.. Local Scalars .. 
  integer :: i, j, xx, yy

  ! Initilize initial condition and RHS

  do j = 0, myData%iRows-1
     do i = 0, myData%iCols -1
        xx = INT(-1.0 + myData%fDx*REAL(i)) ! -1 < x < 1
        yy = INT(-1.0 + myData%fDy*REAL(j)) ! -1 < y < 1
        myData%afU(i, j) = 0.0
        myData%afF(i, j) = - myData%fAlpha * (1.0 - REAL(xx*xx))  &
             * (1.0 - REAL(yy*yy)) - 2.0 * (1.0 - REAL(xx*xx)) &
             - 2.0 * (1.0 - REAL(yy*yy))
     end do
  end do
end subroutine InitializeMatrix

subroutine Finish(myData)
  use VariableDef
  implicit none

  integer :: iErr
  type(JacobiData), intent(inout) :: myData

  deallocate (myData%afU)
  deallocate (myData%afF)

end subroutine Finish

subroutine PrintResults(myData)
  use VariableDef
  implicit none

  type(JacobiData), intent(inout) :: myData

  write (*,328) " Number of iterations : ", myData%iIterCount
  write (*,*)   " Residual             : ", myData%fResidual
  write (*,329) " Solution Error       : ", myData%fError
  write (*,329) " Elapsed Time         : ", myData%fTimeStop - &
       myData%fTimeStart
  write (*,329) " MFlops               : ", 0.000013 * REAL (myData%iIterCount &
       * (myData%iCols - 2) * (myData%iRows - 2))                       &
       / (myData%fTimeStop - myData%fTimeStart)
328 format (A22, I10)
329 format (A22, F16.10)
end subroutine PrintResults


subroutine CheckError(myData)
  use VariableDef
  implicit none

  type(JacobiData), intent(inout) :: myData
  !.. Local Scalars .. 
  integer :: i, j, iErr
  real :: error, temp, xx, yy
  ! ... Executable Statements ...
  error = 0.0

  do j = 0, myData%iRows-1
     do i = 0, myData%iCols -1
        xx = -1.0 + myData%fDx * REAL(i)
        yy = -1.0 + myData%fDy * REAL(j)
        temp = myData%afU(i, j) - (1.0-xx*xx)*(1.0-yy*yy)
        error = error + temp*temp
     end do
  end do

  myData%fError = sqrt(error) / REAL(myData%iCols * myData%iRows)

end subroutine CheckError

